@@ -19,10 +19,12 @@
 
 #include "config.h"
 
+#include "argv.h"
 #include "clipboard.h"
 #include "common/display.h"
 #include "input.h"
 #include "user.h"
+#include "pipe.h"
 #include "sftp.h"
 #include "ssh.h"
 #include "settings.h"
@@ -72,6 +74,7 @@ int guac_ssh_user_join_handler(guac_user* user, int argc, char** argv) {
     /* If not owner, synchronize with current display */
     else {
         guac_terminal_dup(ssh_client->term, user, user->socket);
+        guac_ssh_send_current_argv(user, ssh_client);
         guac_socket_flush(user->socket);
     }
 
@@ -83,6 +86,12 @@ int guac_ssh_user_join_handler(guac_user* user, int argc, char** argv) {
         user->mouse_handler     = guac_ssh_user_mouse_handler;
         user->clipboard_handler = guac_ssh_clipboard_handler;
 
+        /* STDIN redirection */
+        user->pipe_handler = guac_ssh_pipe_handler;
+
+        /* Updates to connection parameters */
+        user->argv_handler = guac_ssh_argv_handler;
+
         /* Display size change events */
         user->size_handler = guac_ssh_user_size_handler;
 
