@@ -67,6 +67,11 @@
  */
 #define GUAC_TELNET_DEFAULT_PASSWORD_REGEX "[Pp]assword:"
 
+/**
+ * The default maximum scrollback size in rows.
+ */
+#define GUAC_TELNET_DEFAULT_MAX_SCROLLBACK 1000
+
 /**
  * Settings for the telnet connection. The values for this structure are parsed
  * from the arguments given during the Guacamole protocol handshake using the
@@ -112,11 +117,30 @@ typedef struct guac_telnet_settings {
      */
     regex_t* password_regex;
 
+    /**
+     * The regular expression to use when searching for whether login was
+     * successful. If no such regex is specified, or if no login failure regex
+     * was specified, this will be NULL.
+     */
+    regex_t* login_success_regex;
+
+    /**
+     * The regular expression to use when searching for whether login failed.
+     * If no such regex is specified, or if no login success regex was
+     * specified, this will be NULL.
+     */
+    regex_t* login_failure_regex;
+
     /**
      * Whether this connection is read-only, and user input should be dropped.
      */
     bool read_only;
 
+    /**
+     * The maximum size of the scrollback buffer in rows.
+     */
+    int max_scrollback;
+
     /**
      * The name of the font to use for display rendering.
      */
@@ -243,6 +267,16 @@ typedef struct guac_telnet_settings {
 guac_telnet_settings* guac_telnet_parse_args(guac_user* user,
         int argc, const char** argv);
 
+/**
+ * Frees the regex pointed to by the given pointer, assigning the value NULL to
+ * that pointer once the regex is freed. If the pointer already contains NULL,
+ * this function has no effect.
+ *
+ * @param regex
+ *     The address of the pointer to the regex that should be freed.
+ */
+void guac_telnet_regex_free(regex_t** regex);
+
 /**
  * Frees the given guac_telnet_settings object, having been previously
  * allocated via guac_telnet_parse_args().
