@@ -19,8 +19,10 @@
 
 #include "config.h"
 
+#include "argv.h"
 #include "clipboard.h"
 #include "input.h"
+#include "pipe.h"
 #include "settings.h"
 #include "telnet.h"
 #include "terminal/terminal.h"
@@ -71,6 +73,7 @@ int guac_telnet_user_join_handler(guac_user* user, int argc, char** argv) {
     /* If not owner, synchronize with current display */
     else {
         guac_terminal_dup(telnet_client->term, user, user->socket);
+        guac_telnet_send_current_argv(user, telnet_client);
         guac_socket_flush(user->socket);
     }
 
@@ -82,6 +85,12 @@ int guac_telnet_user_join_handler(guac_user* user, int argc, char** argv) {
         user->mouse_handler     = guac_telnet_user_mouse_handler;
         user->clipboard_handler = guac_telnet_clipboard_handler;
 
+        /* STDIN redirection */
+        user->pipe_handler = guac_telnet_pipe_handler;
+
+        /* Updates to connection parameters */
+        user->argv_handler = guac_telnet_argv_handler;
+
         /* Display size change events */
         user->size_handler = guac_telnet_user_size_handler;
 
